#! /usr/bin/perl
# Copyright (C) 2001-2002 by <Christian.Queinnec@lip6.fr>
# $Id: FOGE.pl,v 1.35 2003/04/28 17:19:19 queinnec Exp $
# Licensed under the same terms as Perl itself.

package FOGE;

# {{{ Documentation

=head1 NAME

FOGE - A module to generate HTTP forms, store answers in databases,
generate the SQL tables and display statistical results.

=for html Download the package from
<a href="http://youpou.lip6.fr/queinnec/Miscellaneous/FOGE/FOGE-latest.tar.gz">
http://youpou.lip6.fr/queinnec/Miscellaneous/FOGE/FOGE-latest.tar.gz </a>

=for text Download the package from
http://youpou.lip6.fr/queinnec/Miscellaneous/FOGE/FOGE-latest.tar.gz

FOGE is a project hosted by freshmeat.net: see
http://freshmeat.net/projects/foge/

=head1 DESCRIPTION

FOGE is a module to conceive questions whose answer may be open
or a fixed choice of menu items. These questions are gathered in
sections and multiple sections may be displayed in a single HTML page.
When the form is filled, answers are stored in a database. The same
CGI script extracts statistical results or detailed information from
the database. The same CGI script may generate the SQL tables to 
create in the database to hold the results. 

I mainly use that package to get feedback from students on a course.
To have the answers in a database allows to perform some correlations
on individual questions. The main interest of the FOGE package is to
allow the form conceiver to maintain a single data structure for all
aspects: create table, generate the HTML form, handle the content of a
filled form, cull answers for statistical results.

=head1 AUTHOR

Christian Queinnec <Christian.Queinnec@lip6.fr>

=head1 SCREENSHOTS

=for html Here are some screenshots, they are only useful for the HTML
version. Their content is in French but the intention is just to show
the look generated by FOGE. The first screen shows a box to hold a
vote key followed by a section and questions with menus:
<p><img src="screen1q.gif" align="center"></p><p>
The next screen displays the statistics gathered on these questions:
<p><img src="screen1a.gif" align="center"></p>

=for html Here are other types of questions and associated statistics
<p><img src="screen2q.gif" align="center"></p>
<p><img src="screen2a.gif" align="center"></p>

=for html Finally, here is the end of the form:
<p><img src="screen3q.gif" align="center"></p>

=for html But best of all is a 
<a href="http://youpou.lip6.fr/cgi-bin/foge-sample.cgi"> 
running example.</a>

=head1 SYNOPSIS

First, we define a sample.cgi file, the CGI script that will be used to
serve and cull forms. This file should be placed where it may be
served (for instance /var/www/cgi-bin/). This file will define the
questions to ask, their properties (open, single/multiple choice,
menu/buttons, etc.), the way they will appear, the way they may be
mapped onto SQL tables, the way they are statistically counted, etc.
Here follow a simple example.

You start with the usual incantations and the appropriate require form:

 use CGI qw(:standard :html3);
 use CGI::Carp qw(fatalsToBrowser);
 require "FOGE.pl";

You may then define the lists that will serve as choices in buttons,
menus, etc. A number of choices are already defined in the FOGE
package.

 my @Figures = ( 0 .. 9);

You define new questions (some are already defined in the FOGE
package) and you add them to the big bag of questions. You may define
a number of questions, you are not oblidge to use them all or in this
order. 

 FOGE::add_questions({
     'figureAsTiny' => {
         'label'   => "Your preferred figure:",
         'kind'    => 'menu',
         'sqltype' => 'smallint',
         'choice'  => \@Figures},
     'figureAsText' => {
         'label'         => "Your preferred figure:",
         'isValidRegexp' => '^\d$'}, # '$ just for fontlock!
     'understood' => {
         'label'  => "Got it ?", 
         'kind'   => 'menu', 
         'sqltype' => 'boolean', 
         'isValid' => 'coerceToSQLBool', 
         'choice' => \@FOGE::YesNo}, 
 }); 

A question is identified by a key (such as figureAsTiny) in the big bag of
questions. The question is a hash taht must containt the text of the question 
(under the 'label' key), the way it will appear (under the 'kind' key), 
the SQL type (by default 'text') (under the optional key 'sqltype').
In case of a menu, a 'choice' should be specified; this is not mandatory 
for the textfield or textarea kind that correspond to open questions.
Some facilities exist to check the answer with a regexp (under the 
'isValidRegexp' key) or a function that converts the answer to some 
specific SQL type (as encoded by Perl). 

Now that we have a set of questions, we may build a section of some
questions. Note that a section may enforce a question as mandatory.  A
section has a title, an associated SQL table and contains some
questions.

 my @questions1 = (
   {'id' => 'figureAsTiny'},
   {'mandatory' => 1, 'id' => 'understood'});
 my $section1 = {
   'title' => "Your preferences",
   'table' => 'preferred_figures',
   'questions' => \@questions1,
 };

We may build other sections. Once they are all created, we may finish
the complete form, saying:

 my @theSections = (
    $section1,
 );
 my $forms = {
    'title' => "Important poll",
    'enquiry' => "preferred-number", # The identifier for this poll
    'introduction' => "This poll tries to figure out what is
                       the average preferred number.",
    'ipMask'       => '.*',    # Accept any IP number.
    'submit'       => 'VOTE!',
    'sets'         => \@theSections,
    # Time management
        # Don't accept answers after that date:
    'formEndTime'   => timegm(0, 0, 23, 11, 2, 2002),    # 11 mar 2002 23h
        # Don't show statistics before that date:
    'statStartTime' => timegm(0, 0, 23, 11, 2, 2002),    # 11 mar 2002 23h
    # SQL data base to be used:
    'dbName'         => 'dbi:Pg:dbname=someDB host=localhost',
    'dbWriterName'   => 'someDBuser',
    'dbWriterPW'     => 'aVerySecretePW',
    'dbReaderName'   => 'somePublicDBuser',
    'dbReaderPW'     => 'aPublicPW',
    # URL password (Statistics may only be seen if knowing this password):
    'URLpw' => 'anotherSecretePW',
 };

Finally, you just call the default URL handler:

 FOGE::default_main($forms);

And you are done! Now, you can direct your browser to the .cgi file
(pay attention that the FOGE.pl file must be accessible and readable),
You cannot however submit an answer since the database is not ready. 
First, create the SQL tables with 
     
 perl ./sample.cgi 'action=genererTablesEnvoyer' > tables.sql

Then feed your SQL database (Postgresql (I did not try any other DB))
with this SQL table. Do not forget to create the user(s) as specified
in the above $forms variable (these are the values of the 'dbReaderName'
to access statistical results and 'dbWriterName' to record filled-in
forms). Now you may sybmit an answer and it should be recorded in the
database. To see the statistical results, direct your browser towards:
http://.../sample.cgi/anotherSecretePW/see

and look. 

=head1 BUGS

At least three exist: there is a strong bias towards French as for
now.  Not all features are currently described nor all desirable
features are implemented (radio-buttons instead of menus for
instance). Should generate the javascript to check mandatory fields
are filled in before submission.

=head1 HOW TO DEFINE A QUESTION

Questions are represented by a hash and are added to the big bag of
questions via the FOGE:add_questions function. This function takes a
hash mapping question identifier towards hashes describing the
question. This hash should contain some mandatory fields:

=over 4

=item * label

'label' is associated to a string to display the question.

=item * kind

'kind' specifies how to display the question in HTML. By
default, 'kind' is equal to 'menu' if there is a 'choice' entry in the
question otherwise it will be equal to 'text'. The possible kinds are:

=over 4

=item * textfield

'textfield' corresponds to an open question whose answer is a
single word or line. Additionally the 'size' key specifies the length
of the textfield while the optional 'initial' key specifies the
initial content of this textfield.

=item * textarea

'textarea' corresponds to a textfield with several lines. The
'rows' key specifies the number of rows while 'columns' specify the
number of columns of the textarea. The optional 'initial' key
specifies the initial content of this textarea.

=item * menu

'menu' corresponds to a choice of one entry among a fix number
of predefined entries. The 'choice' key is mandatory and specifies the
list of possible answers. The optional 'initial' key specifies the
default choice.

=item * double_menu

'double_menu' is a specific question that corresponds to a
double answer. For instance ``What do you think of ... then'' and
``What do you think of ...(same thing)... now''. The 'choice' key is
mandatory and specifies the list of possible answers. The optional
'initial' key specifies the default choice.

=item * menu_twice

'menu_twice' is a question that expects a double answer. Each of these
answers is obtained via a different menu. The keys are suffixed with 1
or 2 depending on the target menu. Thus the 'choice1' key is mandatory
and specifies the list of possible answers for the first menu. The
optional 'initial1' key specifies the default choice. Keys 'choice2'
and 'initial2' concern the second menu.

=item *  scroll

'scroll' is similar to 'menu' except that multiple answers are
allowed. The 'choice' key is mandatory and specifies the list of
possible answers. The optional 'initial' key specifies the default
choice.

=item * checkbox

'checkbox' is similar to 'scroll' except for the appearance. Instead
of a menu, the choice appears as a row of checkboxes. By default,
checkboxes support multiple selections. If only one is required, set
the 'multiple' option to 0 and the checkboxes will appear as
radio-buttons. The optional 'rows' and 'columns' key may customize the
appearance.

=item * scroll_plus_other

'scroll_plus_other' is similar to 'choice' except that an additional
'textarea' is provided for additional remarks. This is particularly
useful when the set of predefined choices might not cover the full
range of possible answers. The 'rows' key specifies the number of rows
of the textarea while 'columns' specify the number of columns of the
textarea. The 'choice' key is mandatory and specifies the list of
possible answers. The optional 'initial' key specifies the default
choice. Additionally the 'vertical' key if bound to true (i.e. 1)
displays the textarea under the scroll list. If multiples answers are
not expected, set 'multiple' to 0. Allowing multiple answers is
mentioned with the 'multipleAnswerPossible' text.  The label of the
additional 'textarea' may be set with 'otherLabel'.

=back

=item * sqltype

'sqltype' corresponds to the SQL type used to record answers. By
default, this is often 'text' which is the sensible default. In some
cases, it might be 'varchar(.)' (see code). If you have a choice with
numbers only, you may specify 'integer' as well.

=item * isValidRegexp

'isValidRegexp' corresponds to a regexp to check whether the
answer is valid. If the answer is not valid, the submission will not
be accepted, the data are not entered in the database and the user
receives a page telling him that he should correct the answer to the
question first.

=item * isValid

'isValid' corresponds to a user-defined predicate that should
check the answer. This predicate is invoked on the answer, its result
is considered as the appropriate answer to insert in the database. If
its result is the empty string, the answer is considered invalid and
the user is asked to reconsider his submission. 

=item * sum

If 'sum' is true (equal to 1 for instance) then the accumulated sum
will be computed in addition to the observed frequencies. The
accumulated sum is computed from left to right. This is useful to
determine the median of a question with a lot of ordered choices. 

For example, if the 'choices' array is (A, B, C), suppose the 20
answers to be (10, 3, 7) then the result of this question will look
like:

  A      B      C
  10/20   3/20   7/20
  10/20  13/20  20/20

=item * average

If 'average' is true (equal to 1 for instance) then the average is
computed and displayed after the array of observed frequency. For the
average to be computed, labels of choices should be numerically
interpretable! Labels with a number (int or float) within them get
this number as weight. 

For example, if the 'choices' array is (<1h, <2h, >=2h), that is (less
than 1 hour, less than 2 hours, more than 2 hours) suppose the 20
answers to be (10, 3, 7) then the result of this question will look
like:

  <1h    <2h    >=2h
  10/20   3/20   7/20
  10/20  13/20  20/20
  Approximative average: 1.00

The average corresponds to (10*1 + 3*2 + 7*2)/20.

FUTURE work: other option may propose pie-charts!

=back

=head1 HOW TO ASK A QUESTION

When questions are defined in the big bag of questions, they may be
asked.  Questions to be asked are gathered in a section. A section is
a hash with several possible keys. You may have more than one section
in a whole form. Here are the keys:

=over 4

=item * title

'title' is the mandatory title (a string) of the section.

=item * prologue

'prologue' is an optional introduction (a string) to the section.

=item * epilogue

'epilogue' is an optional conclusion (a string) to the section.

=item * questions

'questions' is the mandatory ordered list of questions of that
section.  The questions are specified in a list of hashes. Any of the
hashes may have the following keys:

=over 4

=item * id

'id' specifies the question to use.

=item * mandatory

'mandatory' specifies whether the question is mandatory or
not. Observe that this is not a property of the question per se but an
external property on how the question is deployed. The associated
value is 0 or 1 (by default 0). 

=item * secrete

'secrete' means that no statistics will be publicly displayed for that
question. Statistics are displayed when asked with a the URLpw (see
below) password. This is used when some questions are asked and only a
limited number of people may see the results.

=item * bof

'bof' (a French word meaning ``do not care''). If that
property is set then a --?-- item is added to the current
choice. Observe that this is not a property of the question per se but
an external property on how the question is deployed. The associated
value is 0 or 1 (by default 0).

=back

=item * table

'table' is the name of the SQL table that will hold the answers
of the questions of that section.

=back

=head1 HOW TO DEFINE A WHOLE FORM

When all sections are ready, you may define a form that is, a hash
with the following keys.

=over 4

=item * title

'title' is the title of the whole form. 

=item * introduction

'introduction' is the text (a string) heading the whole form (it
generally explains for whom is this form and what is the intention of
the form). This parameter is optional.

=item * conclusion

'conclusion' is the text (a string) suffixing the whole form (it
generally thank the user that fills the form). This parameter is optional.

=item * ipMask

'ipMask' is a regexp that filters the IP numbers that might
submit answers to this form. This parameter is optional.

=item * submit

'submit' is a word like Submit or VOTE! or Go that appears as
a button to submit an answer to the form.

=item * sets

'sets' specifies the list of sections. This parameter is mandatory.

=item * dbName

'dbName' specifies the database to use. It should follow DBI
conventions.

=item * dbReaderName

'dbReaderName' specifies the database user name to use when extracting
statistics. It should have SQL select access to the tables.

=item * dbReaderPW

'dbReaderPW' is the associated database password for the previous user.

=item * dbWriterName

'dbWriterName' specifies the database user name to use when inserting
new data. It should have SQL insert access to the tables.

=item * dbWriterPW

'dbWriterPW' is the associated database password for the previous user.

=item * dbAnswerTable

'dbAnswerTable' is the name of the table where all answers are
registered. By default, this table is named 'answer'. This is a very
simple table that records any answer, its originating IP number and
date. It confers to the answer a unique ID that is used in all other
tables to identify answers.

=item * URLpw

'URLpw' is a password that must appear on the URL to get statistics.
If left empty, anybody may look at the statistical results of the
form. The knowledge of the URLpw allows to see all statistics even the
ones that concern secrete questions and even before the statistics go
public.

=item * dbKeyTable

Sometimes, it is useful to prevent users to answer more than once to a
form and non-user to ever vote. In that case, specify a 'dbKeyTable'
entry that names a SQL table. This table should be filled with
keys. The form will be asked with an initial textfield asking for a
key. The table will be altered to record that the key is legal and had
been used. The table should have at least three columns (a key, the
name of an enquiry and an identifier (to match the answer in the
answer table)).

The textfield will preceded by the string specified by 'keyPrologue'
and followed by 'keyEpilogue'. The value sent by this textfield may be
set by 'keyName' (or 'dbKeyTable' if absent).

=item * year

'year' specifies the year of the records to consider when
computing statistics.

=item * ignoreLag

'ignoreLag' specifies a number of seconds. When two answers come
from the same IP number separated by less than this number of seconds,
the first one is ignored when computing statistics.

=item * pageFooter

'pageFooter' is the converse of 'introduction'. This is a text
(a string) that appears at the end of the form.

=item * css

'css' is the url of the CSS file to use.

=item * attentionColor

'attentionColor' is the color (by default red) used to display
text requiring some attention.

=item * 

'tableAttributes', 'td1Attributes', 'td2Attributes', 'td3Attributes',
'beginTable', 'endTable' are html/css attributes used for the html
generation. They may be used locally to a section or globally. 

=back

This is the end of this terse documentation.

=cut

# }}}

#use strict;
use vars qw($VERSION);
use diagnostics;
diagnostics::enable();
use Time::Local;
use CGI qw(:standard :html3);
use CGI::Carp qw(fatalsToBrowser);
use DBI;

# {{{ General variables 

my $version = 
    ' $Id: FOGE.pl,v 1.35 2003/04/28 17:19:19 queinnec Exp $ ';
($VERSION) = '$Revision: 1.35 $ ' =~ /\$Revision:\s+([^\s]+)/;

# References of the database used to record the results of the form.
# All these parameters may be customized via the argument to the
# serve() function.

$dbName = 'dbi:Pg:dbname=* host=localhost';
$dbReaderName = $dbWriterName = 'nobody';
$dbReadPW = $dbWriterPW = 'secret';
$dbTrace = 0;    # See DB activity (within /var/log/http/error_log):

# Show the statistics of all questions including the secrete ones.
$showEverything = 0;

# Global parameters for html page generation. All these parameters may
# be customized via the argument to the serve() function.

$pageTitle = "Questionnaire";
$pageFooter = "";
$pageAuthor = 'Christian.Queinnec@lip6.fr';
$pageBGColor = "white";
$attentionColor = "red";
$ipMask = ".*";    # Accept any IP number.
$nowLabel = ' et maintenant: ';
$thenLabel = 'Sur le coup: ';

# Some HTML CSS style settings.

$tableAttributes = "align=\"center\" border=0";
$td1Attributes = "align=\"right\" valign=\"top\" width='50%'";
$td2Attributes = "align=\"center\" valign=\"top\" ";
$td3Attributes = "align=\"left\" valign=\"top\" ";

$beginTable = "<table $tableAttributes>";
$endTable = "</table>";
$bof = "-?-";

# Build a new array with Bof in first position.

sub addBof {
    my($array) = @_;
    my @c = ();
    foreach my $item (@{$array}) {
        push(@c, ($item));
    }
    # Add $bof in first position:
    unshift(@c, ($bof));
    return \@c;
}

sub mayAddBof {
    my($allowBof, $array) = @_;
    if ( $allowBof ) {
        return addBof($array);
    } else {
        return $array;
    }
}

# Determine a property from a question or its section or the whole form:

sub getProperty {
    my($obj, $propertyName) = @_;
    my $result = $obj->{$propertyName};
    if ( defined($result) ) {
        return $result;
    } else {
        $obj = $obj->{'parent'};
        if ( defined($obj) ) {
            return getProperty($obj, $propertyName);
        } else {
            return $obj; # that is, undef
        }
    }
}

sub coerceToSQLBool {
    local($_) = @_;
    if ( /^(Oui|Yes)$/ ) {
        return "True";
    } else {
        return "False";
    }
}

# Should be loaded if necessary via the form.cgi file:                  FUTURE
# For now, we just load it if present or ignore it. This is needed to let 
# old forms run without problem.
eval { require "questions.pl"; };
print STDERR "Warning: Missing questions.pl\n" if $@;

# Add some questions to the predefined bag of questions. The argument
# should have the same structure as bagOfQuestions.

sub add_questions {
    my ($questions) = @_;
    foreach my $key (keys %{$questions}) {
        #print STDERR "*** adding $key\n";
        $bagOfQuestions->{$key} = $questions->{$key};
    }
}

# It is often useful to define a question with a new key but with the
# same attributes as an old question.

sub clone_question {
    my ($key) = @_;
    my $q = $bagOfQuestions->{$key};
    if ( defined($q) ) {
        return $q;
    } else {
        die "No such question $key";
    }
}

# }}}
# {{{ DB Utilities

sub open_db_connection {
    my ($forms, $autocommit) = @_;
    # Assign global variables:
    $dbName = $forms->{'dbName'} if defined($forms->{'dbName'});
    my ($dbUserName, $dbUserPW);
    if ( $autocommit ) {
        # Read the DB:
        if ( defined($forms->{'dbReaderName'}) ) {
            $dbUserName = $forms->{'dbReaderName'};
        } else {
            die "No DB user to read the DB";
        } 
        if ( defined($forms->{'dbReaderPW'}) ) {
            $dbUserPW = $forms->{'dbReaderPW'};
        } else {
            die "No DB user's password to read the DB";
        }
    } else {
        # Modify the DB:
        if ( defined($forms->{'dbWriterName'}) ) {
            $dbUserName = $forms->{'dbWriterName'};
        } else {
            die "No DB user to update the DB";
        } 
        if ( defined($forms->{'dbWriterPW'}) ) {
            $dbUserPW = $forms->{'dbWriterPW'};
        } else {
            die "No DB user's password to update the DB";
        }
    }
    $dbTrace = $forms->{'dbTrace'} if defined($forms->{'dbTrace'});
    my $dbh = DBI->connect($dbName, $dbUserName, $dbUserPW, 
                           { AutoCommit => $autocommit })
        || die "Cannot connect: " . $DBI::errstr;
    $dbh->trace($dbTrace) if $dbTrace;
    return $dbh;
}

sub open_db_connection_for_read {
    my ($forms) = @_;
    open_db_connection($forms, 1);
}
sub open_db_connection_for_update {
    my ($forms) = @_;
    open_db_connection($forms, 0);
}

# }}}
# {{{ Serve form

sub convert_to_text {
    my ($t) = @_;
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdt) = localtime($t);
    my $d = ('dimanche','lundi','mardi','mercredi',
             'jeudi','vendredi','samedi')[$wday];
    my $m = ('janvier','fevrier','mars','avril','mai','juin','juillet',
             'aout','septembre','octobre','novembre','decembre')[$mon];
    my $y = 1900 + $year;
    return "$d $mday $m $y $hour:$min";
}

sub serve_form {
    my ($forms) = @_;
    set_html_parameters($forms);
    my $sections = $forms->{'sets'};

    # Restrict access in IP and time:
    $ipMask = $forms->{'ipMask'} if defined($forms->{'ipMask'});
    my $ipNumber = remote_addr();
    if ( ! ($ipNumber =~ $ipMask) ) {
        die "Not an authorized computer: $ipNumber";
    };
    my $t = time();
    my $startTime = $forms->{'formStartTime'};
    if ( defined($startTime) && $t < $startTime ) {
        die "Not before " . convert_to_text($startTime);
    };
    my $endTime = $forms->{'formEndTime'};
    if ( defined($endTime) && $endTime < $t ) {
        die "Not after " . convert_to_text($endTime);
    };

    print
        header(-Cost => "FRF 0.00312",
               -expires => '+2h' ),
        start_html(-title => $forms->{'title'},
                   -author => $forms->{'author'},
                   -style => {'src'=>$forms->{'css'}} ),
        start_form('post', self_url() . "?t=$t"),
        h1($pageTitle), p, $forms->{'introduction'};

    # Generate a box for a key if the form is a vote.
    my $keyTable = $forms->{'dbKeyTable'};
    if ( defined($keyTable) ) {
        my $keyName = $forms->{'keyName'};
        $keyName = $keyTable unless defined($keyName);
        my $prologue = $forms->{'keyPrologue'};
        $prologue = '' unless defined($prologue);
        my $epilogue = $forms->{'keyEpilogue'};
        $epilogue = '' unless defined($epilogue);
        print 
            $prologue,
            textfield(-name => $keyName,
                      -default => '????-?????',     # should be computed FIXME
                      -size => 10),                 # ditto              FIXME
            $epilogue,
            "\n";
    };

    foreach my $section (@{$sections}) {
        my $prologue = $section->{'prologue'};
        $prologue = '' unless defined($prologue);
        my $epilogue = $section->{'epilogue'};
        $epilogue = '' unless defined($epilogue);
        # shadow global HTML settings                                    FIXME
        local $beginTable = $beginTable;
        local $tableAttributes = $tableAttributes;
        local $td1Attributes = $td1Attributes;
        local $td2Attributes = $td2Attributes;
        local $td3Attributes = $td3Attributes;
        local $endTable = $endTable;
        $beginTable = $section->{'beginTable'} 
        if defined($section->{'beginTable'});
        $tableAttributes = $section->{'tableAttributes'} 
        if defined($section->{'tableAttributes'});
        $td1Attributes = $section->{'td1Attributes'} 
        if defined($section->{'td1Attributes'});
        $td2Attributes = $section->{'td2Attributes'} 
        if defined($section->{'td2Attributes'});
        $td3Attributes = $section->{'td3Attributes'} 
        if defined($section->{'td3Attributes'});
        $endTable = $section->{'endTable'} 
        if defined($section->{'endTable'});
        print
            h2($section->{'title'}), "\n", p,
            $prologue, p,
            $beginTable,
            serve_all_questions($section),
            $endTable, p,
            $epilogue;
    }

    my $conclusion = $forms->{'conclusion'};
    $conclusion = '' unless defined($conclusion);
    print 
        p, hr, "<center>",
        submit(-name => 'action',
               -value => $forms->{'submit'}),
        "</center>", endform,
        p, $conclusion,
        p, hr, "<tt>$version</tt>",
        $pageFooter,
        end_html, "\n";
}

# Convert to HTML all the questions given in arguments.

sub serve_all_questions {
    my ($section) = @_;
    my $questions = $section->{'questions'};
    my $result = "";
    foreach my $q (@{$questions}) {
        my $id = $q->{'id'};
        die "Question without id" unless defined($id);
        my $question = $bagOfQuestions->{$id};
        die "No question named $id" unless defined($question);
        my $kind = $question->{'kind'};
        die "Question $id without kind" unless defined($kind);
        # Create a back pointer from a question to the section containing it:
        $q->{'parent'} = $section;
        my $function = "serve_question_$kind";
        $result .= &$function($q);
    }
    return $result;
}

#    {{{ Methods
# Convert a question into an HTML textfield.

sub serve_question_textfield {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $size = $question->{'size'};
    my $initial = $question->{'initial'};
    $initial = '' unless defined($initial);
    return ligne($label, 
                 $mandatory,
                 textfield(-name => $id,
                           -default => $initial,
                           -size => $size));
}

sub serve_question_textarea {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $rows = $question->{'rows'};
    my $columns = $question->{'columns'};
    my $initial = $question->{'initial'};
    $initial = '' unless defined($initial);
    return ligne($label,  
                 $mandatory,
                 textarea(-name => $id,
                          -rows => $rows,
                          -default => $initial,
                          -columns => $columns));
}

# Convert a question into an HTML menu.

sub serve_question_menu {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $allowBof = getProperty($q, 'bof');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $question->{'choice'};
    $choice = mayAddBof($allowBof, $choice);
    my $initial = $question->{'initial'};
    $initial = '' unless defined($initial);
    return ligne($label, 
                 $mandatory,
                 popup_menu(-name => $id,
                            -default => $initial,
                            -values => $choice));
}

# Convert a question into two html related menus. In fact, the same
# question is asked twice and two answers are expected. This is useful
# when asking a student if he understood some concept at the time of
# the lecture or by the end of the term.

sub serve_question_double_menu {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $then = $q->{'then'};
    $then = $thenLabel unless defined($then);
    my $now = $q->{'now'};
    $now = $nowLabel unless defined($now);
    my $mandatory = getProperty($q, 'mandatory');
    my $allowBof = getProperty($q, 'bof');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $q->{'choice'};
    $choice = $question->{'choice'} if !defined($choice);
    $choice = mayAddBof($allowBof, $choice);
    my $initial = $question->{'initial'};
    $initial = '' unless defined($initial);
    return ligne($label, 
                 $mandatory,
                 strong($then) .
                 popup_menu(-name => "then$id",
                            -default => $initial,
                            -values => $choice) .
                 strong($now) .
                 popup_menu(-name => "now$id",
                            -default => $initial,
                            -values => $choice));
}

# Convert a question asking for two different kinds of answers (as
# specified by two different menus).
# FUTURE extend to n different menus                               FUTURE

sub serve_question_menu_twice {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $allowBof = getProperty($q, 'bof');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice1 = $q->{'choice1'};
    $choice1 = $question->{'choice1'} if !defined($choice1);
    $choice1 = mayAddBof($allowBof, $choice1);
    my $initial1 = $question->{'initial1'};
    $initial1 = '' unless defined($initial1);
    my $choice2 = $q->{'choice2'};
    $choice2 = $question->{'choice2'} if !defined($choice2);
    $choice2 = mayAddBof($allowBof, $choice2);
    my $initial2 = $question->{'initial2'};
    $initial2 = '' unless defined($initial2);
    return ligne($label, 
                 $mandatory,
                 popup_menu(-name => "${id}1",
                            -default => $initial1,
                            -values => $choice1) .
                 popup_menu(-name => "${id}2",
                            -default => $initial2,
                            -values => $choice2));
}

# Convert a question into an HTML menu with multiple possible answers.

sub serve_question_scroll {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $q->{'choice'};
    $choice = $question->{'choice'} if !defined($choice);
    my $initial = $question->{'initial'};
    $initial = '' unless defined($initial);
    return ligne($label, 
                 $mandatory,
                 scrolling_list(-name => $id,
                                -default => $initial,
                                -multiple => 1,
                                -values => $choice));
}

sub serve_question_checkbox {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $q->{'choice'};
    $choice = $question->{'choice'} if !defined($choice);
    my $rows = $q->{'rows'};
    $rows = 1 if !defined($rows);
    my $columns = $q->{'columns'};
    $columns = 1+$#{@{$choice}} if !defined($columns);
    $multiple = $question->{'multiple'};
    $multiple = 1 if !defined($multiple);
    if ( $multiple ) {
        return ligne($label, 
                     $mandatory,
                     checkbox_group(-name => $id,
                                    -rows => $rows,
                                    -columns => $columns,
                                    -values => $choice));
    } else {
        return ligne($label,
                     $mandatory,
                     radio_group(-name => $id,
                                 -rows => $rows,
                                 -default => '-', # deselect all buttons
                                 -columns => $columns,
                                 -values => $choice));
    }
}

sub serve_question_scroll_plus_other {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $question->{'choice'};
    my $initial = $question->{'initial'};
    $initial = '' unless defined($initial);
    my $rows = $question->{'rows'};
    my $columns = $question->{'columns'};
    my $multiple = $question->{'multiple'};
    $multiple = 1 unless defined($multiple);
    my $breakLine = '';
    $breakLine = '</tr><tr>' if $question->{'vertical'};
    my $sl = scrolling_list(-name => $id,
                            -default => $initial,
                            -multiple => $multiple,
                            -values => $choice);
    my $ta = textarea(-name => "other$id",
                      -rows => $rows,
                      -columns => $columns);
    my $multipleMessage = '';
    if ( $multiple ) {
        $multipleMessage = $question->{'multipleAnswerPossible'};
        $multipleMessage = "<em>NOTA: Multiples rponses possibles</em>" 
            unless defined($multipleMessage);
    };
    my $otherLabel = $question->{'otherLabel'};
    $otherLabel = "Autres rponses: " unless defined($otherLabel);
    my $t = "
<table><tr><td>$sl</td>$breakLine
           <td><table><tr><td>$otherLabel</td></tr>
                      <tr><td>$ta</td></tr>
                      <tr><td>$multipleMessage</td></tr>
               </table></td></tr></table>";
    return ligne($label, $mandatory, $t);
}

#    }}}
# }}}
# {{{ HTML Utilities
# Align questions for aesthetics.

sub ligne {
    my($question, $mandatory, $answer) = @_;
    my $mark = mark_mandatory($mandatory);
    return "
<tr><td $td1Attributes> $question</td>
    <td $td2Attributes> $mark</td>
    <td $td3Attributes> $answer</td></tr>
";
}

# Mark mandatory questions with a red star.

sub mark_mandatory {
    my($mandatory) = @_;
    if ( $mandatory ) {
        return "<font color=\"$attentionColor\">*</font>";
    } else {
        return '';
    }
}

# string -> string
# Trim whitespaces around a string.

sub trim {
    local($_) = @_;
    s/^ *(.*?) *$/$1/;
    return $_;
}

# Set global parameters available for the whole HTML page:

sub set_html_parameters {
    my ($forms) = @_;
    $pageTitle = $forms->{'title'} if defined($forms->{'title'});
    $pageFooter = $forms->{'pageFooter'} if defined($forms->{'pageFooter'});
    $pageAuthor = $forms->{'pageAuthor'} if defined($forms->{'pageAuthor'});
    $attentionColor = $forms->{'attentionColor'} 
          if defined($forms->{'attentionColor'});
    $tableAttributes = $forms->{'tableAttributes'} 
          if defined($forms->{'tableAttributes'});
    $td1Attributes = $forms->{'td1Attributes'} 
          if defined($forms->{'td1Attributes'});
    $td2Attributes = $forms->{'td2Attributes'} 
          if defined($forms->{'td2Attributes'});
    $td3Attributes = $forms->{'td3Attributes'} 
          if defined($forms->{'td3Attributes'});
    $beginTable = $forms->{'beginTable'} if defined($forms->{'beginTable'});
    $endTable = $forms->{'endTable'} if defined($forms->{'endTable'});
    # Initialize also global variables
    $bof = $forms->{'bof'} if defined($forms->{'bof'});
}

# }}}
# {{{ Handle a form
# Process a POST request, check mandatory fields to be present, 
# store valid submissions into a table.

sub handle_submission {
    my ($forms) = @_;
    set_html_parameters($forms);
    my $sections = $forms->{'sets'};

    # Restrict access in IP and time:
    $ipMask = $forms->{'ipMask'} if defined($forms->{'ipMask'});
    my $ipNumber = remote_addr();
    if ( ! ($ipNumber =~ $ipMask) ) {
        die "Not an authorized computer: $ipNumber";
    };
    my $t = time();
    my $startTime = $forms->{'formStartTime'};
    if ( defined($startTime) && $t < $startTime ) {
        die "Too early!"
    };
    my $endTime = $forms->{'formEndTime'};
    if ( defined($endTime) && $endTime < $t ) {
        die "Too late!";
    };

    my $id = $t;
    my $enquiry = $forms->{'enquiry'};
    die "No enquiry name" unless defined($enquiry);
    my $allIsOK = 1;
    my $dbh = open_db_connection_for_update($forms);
    # Check if this is a legal vote.
    #### NOTA: may use https to restrict votes submitted by a same person ???
    my $keyTable = $forms->{'dbKeyTable'};
    if ( defined($keyTable) ) {
        my $request = "update $keyTable
set id = ?
where votekey = ?
  and enquiry = '$enquiry'
  and id is null
;";
        my $sth = $dbh->prepare($request)
            || die "Cannot prepare insertkey request: " . DBI::errstr;
        my $keyName = $forms->{'keyName'};
        $keyName = $keyTable unless defined($keyName);
        my $count = $sth->execute($id, param($keyName))
            || die "Cannot execute insertkey request: " . DBI::errstr . "
Auriez-vous omis d'indiquer une cl ?
Auriez-vous commis une erreur en recopiant la cl ?
N'auriez-vous pas dj rpondu au questionnaire ? ";
        die "Clef " . param($keyName) . " non trouve ou usage." 
            if $count == 0;
        $sth->finish();
    };
    my $answerTable = $forms->{'dbAnswerTable'};
    $answerTable = 'answer' if ! defined($answerTable);
    my $who;
    if ( CGI::https() ) {
        $who = CGI::remote_user();
    };
    # Record the global characteristics of the answer:
    my $request = "insert into $answerTable(id, ip, name, remote_user) 
values(?, ?, ?, ?);";
    my $sth = $dbh->prepare($request)
        || die "Cannot prepare insert0 request: " . DBI::errstr;
    $sth->execute($id, $ipNumber, $enquiry, $who)
        || die "Cannot execute insert0 request: " . DBI::errstr 
                . "This may be due to a temporary glitch, please try again.";
    $sth->finish();
    # Record every form fields:
    foreach my $section (@{$sections}) {
        # Stop at first error:
        last if ( ! $allIsOK );
        my $questions = $section->{'questions'};
        my $table = $section->{'table'};
        $request = "insert into $table(id";
        foreach my $q (@{$questions}) {
            my @names = generate_field_names($q);
            foreach my $name (@names) {
                $request .= ", $name";
            }
        }
        $request .= ") values(?";
        $separator = '';
        foreach my $q (@{$questions}) {
            my @names = generate_field_names($q);
            foreach my $name (@names) {
                $request .= ", ?";
            }
        }
        $request .= ");";
        $sth = $dbh->prepare($request)
            || die "Cannot prepare insert request: " . DBI::errstr;
        local $i = 1; # Used by handle_submission_* functions.
        my $ok = '';
        $sth->bind_param($i++, "$id");
        foreach my $q (@{$questions}) {
            my $id = $q->{'id'};
            my $question = $bagOfQuestions->{$id};
            my $kind = $question->{'kind'};
            $kind = 'menu' if (    ! defined($kind) 
                                && defined($question->{'choice'}) );
            $kind = 'text' if ! defined($kind);
            # Create a back pointer from a question to the section containing it:
            $q->{'parent'} = $section;
            my $function = "handle_submission_$kind";
            $ok .= &$function($q, $sth);
        }
        if ( $ok eq '' ) {
            $sth->execute()
                || die "Cannot execute insert1 request: " . DBI::errstr;
        } else {
            $allIsOK = 0;
            print 
                header(-Cost => "FRF 0.00312",
                       -expires => '+1h' ),
                start_html(-title => $forms->{'title'},
                           -author => $forms->{'author'},
                           -style => {'src'=>$forms->{'css'}} ),
                h1("Erreur "), p, "<ul>", 
                $ok, "</ul>", 
                p, hr, "<tt>$version</tt>",
                $pageFooter,
                end_html, "\n";
        }
        $sth->finish;
    }
    if ( $allIsOK ) {
        $dbh->commit()
            || die "Cannot commit insertions: " . DBI::errstr;
        print 
            header(-Cost => "FRF 0.00312",
                   -expires => '+1h' ),
            start_html(-title => $forms->{'title'},
                       -author => $forms->{'author'},
                       -style => {'src'=>$forms->{'css'}} ),
            $forms->{'acknowledgment'},
            p, hr, "<tt>$version</tt>",
            $pageFooter,
            end_html, "\n";
    } else {
        $dbh->rollback;
    }
    $dbh->disconnect;
}

# Stuff a prepared statement with the input values.

sub handle_submission_default {
    my ($q, $sth) = @_;
    my $ok = '';
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my @names = generate_field_names($q);
    foreach my $name (@names) {
        my $v = param($name); # a single value
        if ( $mandatory ) {
            if ( defined($v) ) {
                $v = trim($v);
                if ( $v eq $bof ) {
                    $v = '';
                } else {
                    # $v != $bof
                    my $regexp = $question->{'isValidRegexp'};
                    if ( defined($regexp) ) {
                        if ( !( $v =~ $regexp ) ) {
                            $v = '';
                            $ok .= "<li> Valeur invalide pour la "
                                . "question   " . $label 
                                . " , revenez en arrire et corrigez!";
                        };
                    };
                    my $predicate = $question->{'isValid'};
                    if ( defined($predicate) ) {
                        my $newv = trim(&$predicate($v));
                        if ( $newv ) {
                            $v = $newv;
                        } else {
                            $v = '';
                            $ok .= "<li> Valeur incorrecte pour la "
                                . "question   " . $label 
                                . " , revenez en arrire et corrigez!";
                        };
                    };
                }
            } else {
                # ! defined($v)
                $v = '';
            }
            if ( $v eq '') {
                $ok .= "<li> Vous n'avez pas donn de rponse  "
                    . "la question  " . $label 
                    . " , revenez en arrire et recommencez!";
            };
        };
        # Use the local $i index:
        $sth->bind_param($i++, $v);
    }
    return $ok;
}

sub handle_submission_textfield {
    my ($q, $sth) = @_;
    return handle_submission_default($q, $sth);
}
sub handle_submission_textarea {
    my ($q, $sth) = @_;
    return handle_submission_default($q, $sth);
}
sub handle_submission_menu {
    my ($q, $sth) = @_;
    return handle_submission_default($q, $sth);
}
sub handle_submission_double_menu {
    my ($q, $sth) = @_;
    return handle_submission_default($q, $sth);
}
sub handle_submission_menu_twice {
    my ($q, $sth) = @_;
    return handle_submission_default($q, $sth);
}
sub handle_submission_scroll {
    my ($q, $sth) = @_;
    my $ok = '';
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my @v = param($id); # multiple values
    my $v = trim(join('|', @v));
    # No mandatory stuff, no isValidRegexp, no isValid, here:
    # Use the local $i index:
    $sth->bind_param($i++, $v);
    return $ok;
}
sub handle_submission_checkbox {
    my ($q, $sth) = @_;
    return handle_submission_scroll($q, $sth);
}
sub handle_submission_scroll_plus_other {
    my ($q, $sth) = @_;
    my $ok = '';
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my @v = param($id); # multiple values
    my $v = trim(join('|', @v));
    # No mandatory stuff, no isValidRegexp, no isValid, here:
    # Use the local $i index:
    $sth->bind_param($i++, $v);
    $v = trim(param("other$id"));
    $sth->bind_param($i++, $v);
    return $ok;
}

# Generate the list of names of the fields to be filled in a table
# given the id of a question. These names depend on the nature of the
# question which may involve more than one answer.

sub generate_field_names {
    my ($q) = @_;
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $kind = $question->{'kind'};
    my $function = "generate_field_names_$kind";
    return &$function($id);
}

sub generate_field_names_default {
    my ($id) = @_;
    return ($id);
}
sub generate_field_names_textfield {
    my ($id) = @_;
    return generate_field_names_default($id);
}
sub generate_field_names_textarea {
    my ($id) = @_;
    return generate_field_names_default($id);
}
sub generate_field_names_menu {
    my ($id) = @_;
    return generate_field_names_default($id);
}
sub generate_field_names_double_menu {
    my ($id) = @_;
    return ("then$id", "now$id");
}
sub generate_field_names_menu_twice {
    my ($id) = @_;
    return ("${id}1", "${id}2");
}
sub generate_field_names_scroll {
    my ($id) = @_;
    return generate_field_names_default($id);
}
sub generate_field_names_checkbox {
    my ($id) = @_;
    return generate_field_names_default($id);
}
sub generate_field_names_scroll_plus_other {
    my ($id) = @_;
    return ("$id", "other$id");
}

# }}}
# {{{ Generate SQL table
# Generate SQL tables for the database.

sub generate_table {
    my ($forms) = @_;
    my $request = '';
    my $sections = $forms->{'sets'};
    my $dbWriter = $forms->{'dbWriterName'};
    my $dbWriterPW = $forms->{'dbWriterPW'};
    my $dbReader = $forms->{'dbReaderName'};
    my $dbReaderPW = $forms->{'dbReaderPW'};
    my $answerTable = $forms->{'dbAnswerTable'};
    $answerTable = 'answer' if ! defined($answerTable);
    print "
create user $dbReader with password '$dbReaderPW';
create user $dbWriter with password '$dbWriterPW';

--drop table $answerTable;
create table $answerTable (
       id   int primary key,
       t    timestamp default current_timestamp,
       ip   inet not null,
       name text not null,
       remote_user text null -- if HTTPS is on
);
revoke all on $answerTable from public;
grant select, insert on $answerTable to $dbWriter;
grant select on $answerTable to $dbReader;
";
    # Generate the table to hold keys for the vote:
    my $keyTable = $forms->{'dbKeyTable'};
    my $enquiry = $forms->{'enquiry'};
    if ( defined($keyTable) ) {
        print "
--drop table $keyTable;
create table $keyTable (
    votekey text,
    enquiry text default '$enquiry',
    id      text default null -- references $answerTable
);
revoke all on $keyTable from public;
grant select, update on $keyTable to $dbWriter;
grant select on $keyTable to $dbReader;
";
    };

    #NOTE: the update right is really necessary because of the foreign 
    #keys from other tables that require this access on the Answer table.
    foreach my $section (@{$sections}) {
        my $questions = $section->{'questions'};
        my $table = $section->{'table'};
        $request .= "
--drop table $table;
create table $table (
";
        foreach my $q (@{$questions}) {
            my $id = $q->{'id'};
            my $question = $bagOfQuestions->{$id};
            my $kind = $question->{'kind'};
            # Create a back pointer from a question to the section containing it:
            $q->{'parent'} = $section;
            my $function = "generate_column_$kind";
            $request .= &$function($q);
        }
        $request .=  "\t-- the serial number that identifies the answer:
\tid  \tint references $answerTable
);
revoke all on $table from public;
grant insert, select on $table to $dbWriter;
grant select on $table to $dbReader;
";
    }
    return $request;
}

sub generate_column_textfield {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'varchar(' . $question->{'size'} . ")";
    };
    my $request = '';
    $request .= "\t$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_textarea {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'text';
    };
    my $request = '';
    $request .= "\t$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_menu {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'text';
    };
    my $request = '';
    $request .= "\t$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_double_menu {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'text';
    };
    my $request = '';
    $request .= "\tthen$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    $request .= "\tnow$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_menu_twice {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'text';
    };
    my $request = '';
    $request .= "\t${id}1 \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    $request .= "\t${id}2 \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_scroll {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'text';
    };
    my $request = '';
    $request .= "\t$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_checkbox {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = 'text';
    my $request = '';
    $request .= "\t$id \t$sqltype";
    $request .= " not null" if $mandatory;
    $request .= ",\n";
    return $request;
}
sub generate_column_scroll_plus_other {
    my ($q) = @_;
    my $mandatory = getProperty($q, 'mandatory');
    my $id = $q->{'id'};
    my $question = $bagOfQuestions->{$id};
    my $sqltype = $question->{'sqltype'};
    if ( ! defined($sqltype) ) {
        $sqltype = 'text';
    };
    my $request = '';
    # Ignore mandatory option:
    $request .= "\t$id \t$sqltype";
    $request .= ",\n";
    $request .= "\tother$id \t$sqltype";
    $request .= ",\n";
    return $request;
}
# FUTURE: separate the user that fills table from the one that
# extracts statistics.

# generate_then_register_tables is dangerous. Better to do it by hand
# ie generate the table with the previous function, connect to the
# database and create the tables using an appropriate db user name.

# register_form records questions and their choices within the db.
# }}}
# {{{ Is form ready ?
# Check wether a form is ready that is, SQL tables do exist in the DB and
# have the right structure.

sub check_tables {
    my ($forms) = @_;
    set_html_parameters($forms);
    my $sections = $forms->{'sets'};
    my $answerTable = $forms->{'dbAnswerTable'};
    $answerTable = 'answer' if ! defined($answerTable);
    my $dbh = open_db_connection_for_read($forms);
    # Check the answer table and its columns:
    my @columns = ('id', 't', 'ip', 'name');
    my $result = ul(li(check_one_table($dbh, $answerTable, \@columns)));
    # check the vote table and its columns
    my $keyTable = $forms->{'dbKeyTable'};
    if ( defined($keyTable) ) {
        @columns = ('votekey', 'enquiry', 'id');
        $result .= ul(li(check_one_table($dbh, $keyTable, \@columns)));
    };
    foreach my $section (@{$sections}) {
        my $questions = $section->{'questions'};
        my $table = $section->{'table'};
        @columns = ();
        foreach my $q (@{$questions}) {
            my $id = $q->{'id'};
            push(@columns, $id);
        }
        $result .= ul(li(check_one_table($dbh, $table, \@columns)));
    }
    $dbh->disconnect;
    my $pageTitle = "Check SQL Tables";
    print 
        header(-Cost => "FRF 0.00312",
               -expires => '+2h' ),
        start_html(-title => $forms->{'title'},
                   -author => $forms->{'author'},
                   -style => {'src'=>$forms->{'css'}} ),
        h1($pageTitle), p, $result,
        p, hr, "<tt>$version</tt>",
        $pageFooter,
        end_html, "\n";
}

sub check_one_table {
    my($dbh, $table, $columns) = @_;
    my $request = "select * from $table;";
    DBI::errstr(''); # Reset errors
    my $sth = $dbh->prepare($request);
    $sth->execute;
    my @result;
    # Report the DB error:
    push(@result, (em(DBI::errstr). "\n")) if defined(DBI::errstr);
    foreach my $c ( @{$columns} ) {
        $c =~tr/A-Z/a-z/;
        # NAME_lc_hash is problematic for old CGI package versions:
        if ( ! defined($sth->{NAME_lc_hash}{$c}) ) {
            # Report missing field:
            push(@result, "Field $c is missing from table $table.");
        };
    }
    my @row;
    my $count = 0;
    while ( @row = $sth->fetchrow_array ) {
        $count++;
    }
    # Report number of records:
    push(@result, "There are $count records in table $table.");
    $sth->finish;
    return \@result;
}
# FUTURE: should also check the type of the column.

# }}}
# {{{ Statistics
# cull, histogram etc 

sub compute_statistics {
    my ($year, $delay, $forms, $checkDate) = @_;
    set_html_parameters($forms);
    my $sections = $forms->{'sets'};
    my $answerTable = $forms->{'dbAnswerTable'};
    $answerTable = 'answer' if ! defined($answerTable);
    my $enquiry = $forms->{'enquiry'};
    die "No enquiry name" if !defined($enquiry);

    # Restrict access in IP and time:
    $ipMask = $forms->{'ipMask'} if defined($forms->{'ipMask'});
    my $ipNumber = remote_addr();
    if ( ! ($ipNumber =~ $ipMask) ) {
        die "Not an authorized computer: $ipNumber";
    };
    my $t = time();
    my $startTime = $forms->{'statStartTime'};
    if ( $checkDate && defined($startTime) && $t < $startTime ) {
        die "Too early!"
    };
    my $endTime = $forms->{'statEndTime'};
    if ( $checkDate && defined($endTime) && $endTime < $t ) {
        die "Too late!";
    };

    my $dbh = open_db_connection_for_read($forms);
    # Filter out records coming from the same IP number and separated by 
    # less than $delay seconds. Only the last one is considered.
    my $sth = $dbh->prepare("
create temporary table IDs as
select A1.id
from $answerTable A1
where A1.name = '$enquiry'
  and date_part('year', A1.t) = $year
  and not exists (
    select *
    from $answerTable A2
    where A2.ip = A1.ip
      and A2.name = '$enquiry'
      and A2.t > A1.t
      and date_part('epoch', A2.t) < date_part('epoch', A1.t) + $delay
                 );
")      || die "Cannot prepare temp table: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request0: " . DBI::errstr;
    $sth->finish;
    # Count the number of answers: 
    $sth = $dbh->prepare("
select count(*) 
from IDs
")      || die "Cannot prepare request1: " . DBI::errstr; 
    $sth->execute()
        || die "Cannot execute request1: " . DBI::errstr;
    my $count = 1;
    if ( @row = $sth->fetchrow_array ) {
        ($count) = @row;
    } else {
        die "Cannot count answers: " . DBI::errstr;
    }
    $sth->finish;

    print
        header(-Cost => "FRF 0.00312",
               -expires => '+2h' ),
        start_html(-title => $forms->{'title'},
                   -author => $forms->{'author'},
                   -style => {'src'=>$forms->{'css'}} ),
        h1($pageTitle), p, "Statistiques pour l'enqute ", 
        tt($forms->{'enquiry'}), ", il y a eu en tout $count rponse",
        (($count>1)?"s":""), ".";

    foreach my $section (@{$sections}) {
        my $prologue = $section->{'prologue'};
        if ( ! defined($prologue) ) {
            $prologue = "";
        };
        my $epilogue = $section->{'epilogue'};
        if ( ! defined($epilogue) ) {
            $epilogue = "";
        };
        print
            h2($section->{'title'}), "\n", p,
            $prologue, p,
            $beginTable,
            compute_statistic_all_questions($dbh, $count, $section),
            $endTable, p,
            $epilogue;
    }

    print 
        p, hr, "<tt>$version</tt>",
        $pageFooter,
        end_html, "\n";
    $dbh->disconnect;
}

sub compute_statistic_all_questions {
    my($dbh, $count, $section) = @_;
    my $questions = $section->{'questions'};
    my $result = "";
    foreach my $q (@{$questions}) {
        my $id = $q->{'id'};
        my $table = $section->{'table'};
        die "Question without id" unless defined($id);
        my $question = $bagOfQuestions->{$id};
        die "No question named $id" unless defined($question);
        my $kind = $question->{'kind'};
        die "Question $id without kind" unless defined($kind);
        # Create a back pointer from a question to the section containing it:
        $q->{'parent'} = $section;
        my $secrete = getProperty($q, 'secrete');
        if ( ! $showEverything && defined($secrete) && $secrete ) {
            # Hide the statistics about this question!
        } else {
            my $function = "count_question_$kind";
            $result .= &$function($dbh, $count, $q, $table);
        }
    }
    return $result;
}

sub count_question_textfield {
    my($dbh, $count, $q, $table) = @_;
    return count_question_textarea($dbh, $count, $q, $table);
}

sub count_question_textarea {
    my($dbh, $count, $q, $table) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $allowBof = getProperty($q, 'bof');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $sth = $dbh->prepare("
select T.$id 
from IDs i, $table T
where T.id = i.id
group by T.$id
")      || die "Cannot prepare request2: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request2: " . DBI::errstr;
    my $results = "<table border=1>\n";
    while ( @row = $sth->fetchrow_array ) {
        my($answer) = @row;
        # Remove bad words ?
        if ( defined($answer) && !( $answer =~ /^\s*$/ ) ) {
            $results .= "<tr><td>$answer</td></tr>";
        };
    }
    $results .= "</table>\n";
    $sth->finish;
    return ligne($label,
                 $mandatory,
                 $results);
}

sub count_question_menu {
    my($dbh, $count, $q, $table) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my @sum = ();
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $q->{'choice'};
    $choice = $question->{'choice'} if !defined($choice);
    my $sth = $dbh->prepare("
select T.$id, count(*) as total
from IDs i, $table T
where T.id = i.id
group by T.$id
")      || die "Cannot prepare request3: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request3: " . DBI::errstr;
    my $results = "<table border=1><tr>\n";
    my %h = ();
    # Fill in the 'h' hashtable:
    while ( @row = $sth->fetchrow_array ) {
        my($answer, $total) = @row;
        $answer = trim("$answer");
        $h{$answer} = $total;
        #print STDERR "**************** A=$answer, T=$total, (", $#h ,"\n";
    }
    # Emit histogram along the order of the 'choice' sequence.
    foreach my $a ( @{$choice} ) {
        next if $a eq $bof;
        $a = trim($a);
        my $total = $h{$a};
        if ( defined($total) ) {
            push(@sum, $total);
            undef $h{$a};
            my $percentile = int((100 * $total) / $count);
            #print STDERR "**************** a=$a, p=$percentile, (", $#h, "\n";
            $results .= "<td>$a: $percentile% ($total/$count)</td>";
        };
    }
    # Handle remaining choices (probably boolean (or encoded) values):
    foreach my $a ( keys %h ) {
        my $total = $h{$a};
        if ( defined($total) && $total > 0 ) {
            my $percentile = int((100 * $total) / $count);
            #print STDERR "****************rest a=$a, p=$percentile\n";
            $results .= "<td>$a: $percentile% 
<small>($total/$count)</small></td>";
        };
    }
    $results .= "</tr>";
    # Compute the accumulated sum if required.
    if ( $question->{'sum'} ) {
        $results .= "\n<tr>";
        my $total = 0;
        foreach my $c ( @sum ) {
            $total += $c;
            $side = 'right';
            if ( 2*$total >= $count ) {
                $side = 'left';
            };
            $results .= "<td align='$side'>
<small>&Sigma;:$total/$count</small></td>"
        }
        $results .= "</tr>\n";
    };
    $results .= "</table>\n";
    # Compute average if required
    if ( $question->{'average'} ) {
        my $average = 0.0;
        my $i = 0;
        foreach my $a ( @{$choice} ) {
            next if $a eq $bof;
            my ($weight) = ( $a =~ /(\d+([\.,]\d*)?)/ );
            if ( defined($weight) && defined($sum[$i]) ) {
                $average += $weight * $sum[$i];
            };
            $i++;
        }
        $average = (int(100 * $average/$count))/100.0;
        $results .= "Moyenne approximative: $average";
    };
    $sth->finish;
    return ligne($label,
                 $mandatory,
                 $results);
}

sub identity {
    my ($x) = @_;
    return $x;
}

sub count_question_double_menu {
    my($dbh, $count, $q, $table) = @_;
    my $id = $q->{'id'};
    my $then = $q->{'then'};
    $then = $thenLabel unless defined($then);
    my $now = $q->{'now'};
    $now = $nowLabel unless defined($now);
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $q->{'choice'};
    $choice = $question->{'choice'} if (!defined($choice));
    my $results = "<table border=1><tr><th></th>\n";
    foreach my $a ( @{$choice} ) {
        next if $a eq $bof;
        $results .= "<th>$a</th>";
    }
    $results .= "</tr>\n<tr><th>$then</th>";
    my $sth = $dbh->prepare("
select T.then$id, count(*) as total
from IDs i, $table T
where T.id = i.id
group by T.then$id
")      || die "Cannot prepare request4: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request4: " . DBI::errstr;
    my %thenResults = ();
    while ( @row = $sth->fetchrow_array ) {
        my($answer, $total) = @row;
        $thenResults{trim($answer)} = $total;
        #print STDERR "A=$answer, T=$total\n";
    }
    $sth->finish;
    foreach my $a ( @{$choice} ) {
        next if $a eq $bof;
        my $total = $thenResults{trim($a)};
        if ( defined($total) ) {
            my $percentile = int((100 * $total) / $count);
            #print STDERR "a=$a, p=$percentile\n";
            $results .= "<td>$percentile% ($total/$count)</td>";
        };
    }
    $sth = $dbh->prepare("
select T.now$id, count(*) as total
from IDs i, $table T
where T.id = i.id
group by T.now$id
")      || die "Cannot prepare request5: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request5: " . DBI::errstr;
    my %nowResults = ();
    while ( @row = $sth->fetchrow_array ) {
        my($answer, $total) = @row;
        $nowResults{trim($answer)} = $total;
        #print STDERR "A=$answer, T=$total\n";
    }
    $sth->finish;
    $results .= "</tr>\n<tr><th>$now</th>";
    foreach my $a ( @{$choice} ) {
        next if $a eq $bof;
        my $total = $nowResults{trim($a)};
        if ( defined($total) ) {
            my $percentile = int((100 * $total) / $count);
            #print STDERR "a=$a, p=$percentile\n";
            $results .= "<td>$percentile% ($total/$count)</td>";
        };
    }
    $results .= "</tr></table>\n";
    return ligne($label,
                 $mandatory,
                 $results);
}

sub count_question_menu_twice {
    my($dbh, $count, $q, $table) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $results = "<table border=1>\n";
    for ( $i=1 ; $i<3 ; $i++ ) {
        my $choice = $q->{"choice$i"};
        $choice = $question->{"choice$i"} if !defined($choice);
        my $sth = $dbh->prepare("
select T.${id}$i, count(*) as total
from IDs i, $table T
where T.id = i.id
group by T.${id}$i
")      || die "Cannot prepare request3t: " . DBI::errstr;
        $sth->execute()
            || die "Cannot execute request3t: " . DBI::errstr;
        $results .= "<tr>\n";
        my %h = ();
        while ( @row = $sth->fetchrow_array ) {
            my($answer, $total) = @row;
            $answer = trim("$answer");
            $h{$answer} = $total;
            #print STDERR "**************** A=$answer, T=$total, (", $#h ,"\n";
        }
        foreach my $a ( @{$choice} ) {
            next if $a eq $bof;
            $a = trim($a);
            my $total = $h{$a};
            if ( defined($total) ) {
                undef $h{$a};
                my $percentile = int((100 * $total) / $count);
                #print STDERR "**************** a=$a, p=$percentile, (", $#h, "\n";
                $results .= "<td>$a: $percentile% ($total/$count)</td>";
            };
        }
        # Handle remaining choices (probably boolean (or encoded) values):
        foreach my $a ( keys %h ) {
            my $total = $h{$a};
            if ( defined($total) && $total > 0 ) {
                my $percentile = int((100 * $total) / $count);
                #print STDERR "****************rest a=$a, p=$percentile\n";
                $results .= "<td>$a: $percentile% ($total/$count)</td>";
            };
        }
        $results .= "</tr>\n";
        $sth->finish;
    }
    $results .= "</table>\n";
    return ligne($label,
                 $mandatory,
                 $results);
}

sub count_question_scroll {
    my($dbh, $count, $q, $table) = @_;
    my $id = $q->{'id'};
    my $mandatory = getProperty($q, 'mandatory');
    my $question = $bagOfQuestions->{$id};
    my $label = $question->{'label'};
    my $choice = $q->{'choice'};
    $choice = $question->{'choice'} if !defined($choice);
    my $sth = $dbh->prepare("
select T.$id
from IDs i, $table T
where T.id = i.id
")      || die "Cannot prepare request2x: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request2x: " . DBI::errstr;
    my $results = "<table border=1>\n";
    my %h = ();
    while ( @row = $sth->fetchrow_array ) {
        my($answer) = @row;
        my @answers = split /\|/, $answer;
        foreach my $a ( @answers) {
            #print STDERR "**************** a=$a\n"; # DEBUG
            $a = trim($a);
            if ( defined($h{$a}) ) {
                $h{$a}++;
            } else {
                $h{$a} = 1;
            }
        }
    }
    foreach my $a ( @{$choice} ) {
        next if $a eq $bof;
        $a = trim($a);
        my $total = $h{$a};
        if ( defined($total) ) {
            undef $h{$a};
            my $percentile = int((100 * $total) / $count);
            #print STDERR "**************** a=$a, p=$percentile, (", $#h, "\n";
            $results .= "<td>$a: $percentile% ($total/$count)</td>";
        };
    }
    # Handle remaining choices (probably boolean (or encoded) values):
    foreach my $a ( keys %h ) {
        my $total = $h{$a};
        if ( defined($total) && $total > 0 ) {
            my $percentile = int((100 * $total) / $count);
            #print STDERR "****************rest a=$a, p=$percentile\n";
            $results .= "<td>$a: $percentile% ($total/$count)</td>";
        };
    }
    $results .= "</tr></table>\n";
    $sth->finish;
    return ligne($label, $mandatory, $results);
}

sub count_question_checkbox {
    my($dbh, $count, $q, $table) = @_;
    return count_question_scroll($dbh, $count, $q, $table);
}

sub count_question_scroll_plus_other {
    my($dbh, $count, $q, $table) = @_;
    my $line1 = count_question_scroll($dbh, $count, $q, $table);
    my $id = $q->{'id'};
    # Handle additional comments 
    my $sth = $dbh->prepare("
select T.other$id 
from IDs i, $table T
where T.id = i.id
")      || die "Cannot prepare request2y: " . DBI::errstr;
    $sth->execute()
        || die "Cannot execute request2y: " . DBI::errstr;
    my $results = "<table border=1>\n";
    while ( @row = $sth->fetchrow_array ) {
        my($answer) = @row;
        # Remove bad words ?
        if ( defined($answer) && !( $answer =~ /^\s*$/ ) ) {
            $results .= "<tr><td>$answer</td></tr>";
        };
    }
    $results .= "</table>\n";
    $sth->finish;
    my $line2 = ligne('', '', $results);
    return $line1 . $line2;
}

# }}}
# {{{ Default main
# FIX French names

sub default_main {
    return if $test_mode;
    my($forms) = @_;
    # The action may be set by a POST or GET parameter:
    if ( defined(param('action')) ) {
        if ( param('action') eq $forms->{'submit'} ) {
            handle_submission($forms);
        } elsif ( param('action') eq 'genererTablesEnvoyer' ) {
            print generate_table($forms);
        } elsif ( param('action') eq 'verifierTables' ) {
            print check_tables($forms);
        } else {
            die "Action not understood: " . param('action');
        }
    } else {
        # No button was pressed, check if the URL contains some special words
        # intended to be used by the teaching team:
        my $pi = path_info();
        path_info("/");
        if ( defined($pi) ) {
            my $password = $forms->{'URLpw'};
            $password = '/' . (defined($password) ? $password : '');
            if ( $pi =~ "$password/see" ) {
                $showEverything = 1;
                my $year = $forms->{'year'};
                $year = 2002 if !defined($year);
                my $ignoreLag = $forms->{'ignoreLag'};
                $ignoreLag = 30 if !defined($ignoreLag);
                compute_statistics($year, $ignoreLag, $forms, 0);
            } elsif ( $pi =~ "$password/check" ) {
                check_tables($forms);
            } elsif ( $pi =~ "$password/serve" ) {
                # Serve the form even if out of time:
                $forms->{'formStartTime'} = undef;
                $forms->{'formEndTime'} = undef;
                serve_form($forms);
            } elsif ( $pi =~ "/see" ) {
                $showEverything = 0;
                my $year = $forms->{'year'};
                $year = 2002 if !defined($year);
                my $ignoreLag = $forms->{'ignoreLag'};
                $ignoreLag = 30 if !defined($ignoreLag);
                compute_statistics($year, $ignoreLag, $forms, 1);
            } else {
                serve_form($forms);
            }
        } else {
            serve_form($forms);
        }
    }
}

# FUTURE: make a page for all the derived services (statistics, check
# tables, generate tables, see created tables, ...) so a click is
# sufficient to get the information.

# }}}

@YesNo = ("Oui", "Non");

1;

# end of FOGE.pl
